/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "log.h"
#include "sysstat.h"

#include <alloca.h>
#include <assert.h>
#ifdef USECURSES
#include <curses.h>
#else
#include <ncurses.h>
#endif
#include <errno.h>
#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <term.h>
#include <unistd.h>


static WINDOW *w = 0, *H = 0;
enum { BLACK_ON_WHITE = 1, BLUE_ON_WHITE, RED_ON_WHITE };
static int Sigwinch = 1, stress, nostress, header, underline, 
    Numlines = 0, Numcol = 0, Oldcurs;
static sysstat_t *displayed = 0;
static clock_t Dis_ts;
extern int Fslines;

void
exit_curses(void)
{
	curs_set(Oldcurs);
	endwin();
}


void
sighandler(int sig)
{
	if (sig == SIGINT || sig == SIGTERM) {
		exit(EXIT_SUCCESS);
	}
	assert(sig == SIGWINCH);
	Sigwinch = 1;
}


void
display_help(void)
{
	char *line;

	if (H == 0) {
		if ((Numlines == 0) || (Numcol == 0)) {
			struct winsize w;
			if (-1 == ioctl(STDOUT_FILENO,TIOCGWINSZ,&w))
				error("cannot read terminal size: %s\n",strerror(errno));
			Numlines = w.ws_row;
			Numcol = w.ws_col;
		}
		assert(Numlines && Numcol);
		H = newwin(Numlines,Numcol,0,0);
		line = alloca(Numcol + 1);
		wclear(H);
		memset(line,'=',Numcol);
		line[Numcol] = 0;
		mvwaddstr(H, 0,0,"sysstat version "VERSION);
		mvwaddstr(H, 1,0,"Copyright 2006, Thomas Maier-Komor");
		mvwaddstr(H, 2,0,"license: GPLv2 - see file COPYING or \"man gpl\" for details");
		mvwaddstr(H, 3,0,line);

		mvwaddstr(H, 5,0,"Hotkeys:                         Field descriptions:");
		mvwaddstr(H, 6,0,"--------                         -------------------");

		mvwaddstr(H, 7,0,"'q' : quit sysstat               LoadAv:    load average over time interval");
		mvwaddstr(H, 8,0,"'?' : toggle help screen         intr:      interrupts");
		mvwaddstr(H, 9,0,"                                 intrblk:   blocked interrupts");
		mvwaddstr(H,10,0,"'h' : display next host          xcalls:    cpu to cpu cross calls");
		mvwaddstr(H,11,0,"'H' : display previous host      pswitch:   process/context switches");
		mvwaddstr(H,12,0,"                                 migrate:   inter cpu migrations of processes");
		mvwaddstr(H,13,0,"sort process list by:            invswtch:  involuntary context switches");
		mvwaddstr(H,14,0,"'u' : username                   scan:      pager deamon's page scans");
		mvwaddstr(H,15,0,"'n' : NPROC                      pgswapin:  pages swapped in");
		mvwaddstr(H,16,0,"'r' : RSS                        pgswapout: pages swapped out");
		mvwaddstr(H,17,0,"'s' : SIZE                       NPROC:     number of processes");
		mvwaddstr(H,18,0,"'+' : H+S                        O/S/R/Z/T: oncpu/sleep/runnable/zompie/stopped");
		mvwaddstr(H,19,0,"'t' : Time                       NLWP:      number of light weight processes");
		mvwaddstr(H,20,0,"'c' : %CPU                       SIZE:      allocated memory");
		mvwaddstr(H,21,0,"                                 RSS:       resident memory");
		mvwaddstr(H,22,0,"                                 H+S:       heap plus stack");
		wsyncup(H);
		wrefresh(H);
	} else {
		delwin(H);
		H = 0;
		refresh();
	}
}


int
init_curses(void)
{
	void (*osig)(int);

	if (0 == isatty(STDOUT_FILENO))
		return 0;
	w = initscr();
	noecho();
	cbreak();
	intrflush(stdscr,FALSE);
	nodelay(w, TRUE);
	nonl();
	Oldcurs = curs_set(0);
	if (has_colors()) {
		start_color();
		init_pair(BLACK_ON_WHITE,COLOR_BLACK,COLOR_WHITE);
		init_pair(BLUE_ON_WHITE,COLOR_BLUE,COLOR_WHITE);
		init_pair(RED_ON_WHITE,COLOR_RED,COLOR_WHITE);
		nostress = COLOR_PAIR(BLACK_ON_WHITE);
		stress = A_BOLD | COLOR_PAIR(RED_ON_WHITE);
		header = COLOR_PAIR(BLUE_ON_WHITE) | A_UNDERLINE;
		underline = COLOR_PAIR(BLACK_ON_WHITE) | A_UNDERLINE;
	} else {
		header = A_UNDERLINE;
		underline = A_UNDERLINE;
		nostress = 0;
		stress = A_BOLD;
	}
	osig = signal(SIGWINCH,sighandler);
	assert(osig != SIG_ERR);
	osig = signal(SIGINT,sighandler);
	assert(osig != SIG_ERR);
	osig = signal(SIGTERM,sighandler);
	assert(osig != SIG_ERR);
	atexit(exit_curses);
	return 1;
}


static char *
per2str(double d)
{
	static char str[7];
	assert(d <= 1E2);
	if (d < 0) {
		memcpy(str,"n/a ",5);
	} else if (d < 2E-2) {
		sprintf(str,"%4.0f",d);
	} else if (d < 9.995) {
		sprintf(str,"%4.2f",d);
	} else if (d < 99.95) {
		sprintf(str,"%4.1f",d);
	} else
		sprintf(str,"%4.0f",rint(d));
	return str;
}


static char *
dbl2str(double d)
{
	static char str[7];
	double r = rint(d);
	if (d < 0) {
		memcpy(str,"unav.",6);
	} else if (d < 2E-2) {
		sprintf(str,"%4.0f ",d);
	} else if (d < 9.995) {
		sprintf(str,"%4.2f ",d);
	} else if (r < 1E2) {
		sprintf(str,"%4.1f ",d);
	} else if (r < 1E4) {
		sprintf(str,"%4.0f ",r);
	} else if (r < 1E5) {
		d /= 1E3;
		sprintf(str,"%4.1fk",d);
	} else if (d < 10E6) {
		d /= 1E3;
		sprintf(str,"%4.0fk",rint(d));
	} else if (d < 100E6) {
		d /= 1E6;
		sprintf(str,"%4.1fM",rint(d));
	} else {
		d /= 1E6;
		sprintf(str,"%4.0fM",rint(d));
	}
	return str;
}


static char *
kb2str(long long ll)
{
	static char str[8];
	char dim = ' ';
	double v;

	if (ll < 0) {
		memcpy(str,"???",4);
		return str;
	} else if (ll >= 10000LL*1024LL*1024LL) {
		v = (double)(ll >> 20)/1024;
		dim = 'T';
	} else if (ll >= 10000*1024) {
		v = (double)(ll >> 10)/1024;
		dim = 'G';
	} else if (ll >= 10000) {
		v = (double)ll/1024;
		dim = 'M';
	} else {
		v = (double) ll;
		dim = 'k';
	}
	if (rint(v) >= 100)
		snprintf(str,sizeof(str),"%4.0f%c",v,dim);
	else
		snprintf(str,sizeof(str),"%4.1f%c",v,dim);
	return str;
}


static char *
dkb2str(double v)
{
	static char str[8];
	char dim = ' ';

	if (v < 0) {
		memcpy(str,"???",4);
		return str;
	} else if (v >= 10000LL*1024LL*1024LL) {
		v /= 1024.0*1024.0*1024.0;
		dim = 'T';
	} else if (v >= 10000*1024.0) {
		v /= 1024*1024.0;
		dim = 'G';
	} else if (v >= 10000) {
		v /= 1024.0;
		dim = 'M';
	} else {
		dim = 'k';
	}
	if (rint(v) >= 100)
		snprintf(str,sizeof(str),"%4.0f%c",v,dim);
	else
		snprintf(str,sizeof(str),"%4.1f%c",v,dim);
	return str;
}


unsigned 
display_curses(sysstat_t *s)
{
	int num;
	userstat_t *u;
	fsstat_t *fs;
	char line[4096], *l;
	int fslines = Fslines, ulines = 0, unshown = 0;
	sysstat_t *h = s;
	clock_t now;

	now = time(0);
	if (H || (s == 0) || (s->TS == 0) ||
	    ((s == displayed) && (s->TS <= Dis_ts) && (now - Dis_ts < 60) && (Sigwinch == 0)))
		return 0;
	displayed = s;
	Dis_ts = s->TS;
	h = s->next;
	while (h != s) {
		++unshown;
		h = h->next;
	}
	assert(w);
	if (Sigwinch) {
		struct winsize ws;
		if (-1 == ioctl(STDOUT_FILENO,TIOCGWINSZ,&ws))
			error("cannot read terminal size: %s\n",strerror(errno));
		LINES = Numlines = ws.ws_row;
		COLS = Numcol = ws.ws_col;
		assert(COLS < sizeof(line));
		dbug("SIGWINCH: Numlines = %d\n",Numlines);
		endwin();
		refresh();
		redrawwin(stdscr);
#ifdef USECURSES
		w = initscr();
#endif
		Sigwinch = 0;
		(void) signal(SIGWINCH, sighandler);
	}
	if (unshown) 
		--Numlines;
	num = 0;
	if (fslines > s->numfs) {
		dbug("fslines = %d, numfs = %d\n",fslines,s->numfs);
		fslines = s->numfs;
	}
	attrset(nostress);
	erase();
	attrset(header);
	memset(line,' ',sizeof(line));
	strcpy(line,s->Hostinfo);
	l = line + strlen(s->Hostinfo);
	if (s->Ocpus == s->Ncpus) {
		if (s->Ncpus > 1) {
			l += sprintf(l,", %d CPUs", s->Ocpus);
		} else {
			assert(s->Ocpus == 1);
			l += sprintf(l,", 1 CPU");
		}
	} else {
		assert(s->Ncpus > 1);
		l += sprintf(l,", %d/%d CPUs",s->Ocpus,s->Ncpus);
	}
	if (s->BootTime) {
		int d,h,m;
		clock_t b = s->TS - s->BootTime;
		d = b / (60*60*24);
		b -= d * (60*60*24);
		h = b / (60*60);
		b -= h * (60*60);
		m = b / (60);
		b -= m * (60);
		if (d == 1)
			l += sprintf(l,", up 1 day");
		else if (d)
			l += sprintf(l,", up %d days",d);
		else
			l += sprintf(l,", up %02d:%02d:%02d",h,m,b);
	} else if (s->Devices != 1) {
		l += sprintf(l,", %lu disks ", s->Devices);
	} else if (s->Devices == 1) {
		l += sprintf(l,", 1 disk ");
	}
	*l = ' ';
	strftime(line+61,21,"%c",localtime(&s->TS));
	if (now - s->TS > 60) {
		Dis_ts = now;
		attrset(stress);
	}
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	attrset(nostress);
	l = line;
	l += sprintf(l," User   CPU: %s%% |", per2str(s->UserP));
	l += sprintf(l," LoadAv/1m:  %s |", dbl2str(s->loadavg[0]));
	l += sprintf(l," Physical:    %s |", kb2str(s->Physmem));
	if (s->IpInR >= 0)
		l += sprintf(l," IP rcv/s:   %s ", dbl2str(s->IpInR));
	else
		l += sprintf(l," ipacket/s:  %s ", dbl2str(s->NetInR));
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	l = line;
	l += sprintf(l," Kernel CPU: %s%% |",per2str(s->KernelP));
	l += sprintf(l," LoadAv/5m:  %s |", dbl2str(s->loadavg[1]));
	l += sprintf(l," Free:        %s |", kb2str(s->Pagesfree));
	if (s->IpOutR >= 0)
		l += sprintf(l," IP send/s:  %s ", dbl2str(s->IpOutR));
	else
		l += sprintf(l," opacket/s:  %s ", dbl2str(s->NetOutR));
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	attrset(underline);
	mvaddstr(num++,0," Idle   CPU: ");
	if (s->IdleP <= 1)
		attrset(stress | A_UNDERLINE);
	sprintf(line,"%s%%",per2str(s->IdleP));
	addstr(line);
	attrset(underline);
	l = line;
	l += sprintf(l," | LoadAv/15m: %s |", dbl2str(s->loadavg[2]));
	l += sprintf(l," Locked:      %s |", kb2str(s->Pageslocked));
	l += sprintf(l," net errs: %6lu  ", s->IpInDiscards+s->IpOutDiscards+s->IpInErrs);
	addstr(line);
	if (num == Numlines) goto done;

	attrset(nostress);
	l = line;
	l += sprintf(l," intr/s:     %s |", dbl2str(s->IntrR));
	l += sprintf(l," syscall/s:  %s |", dbl2str(s->SyscallR));
	l += sprintf(l," SwapTotal:   %s |", kb2str(s->SwapTotal));
	l += sprintf(l," nfsd ops/s: %s ", dbl2str(s->NfsdR));
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	l = line;
	l += sprintf(l," intrblk/s:  %s |", dbl2str(s->IntrBlkR));
	l += sprintf(l," sysfork/s:  %s |", dbl2str(s->SysforkR));
	l += sprintf(l," SwapFree:    %s |", kb2str(s->SwapFree));
	l += sprintf(l," NfsIn/s:    %s ", dkb2str(s->NfsIR));
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	l = line;
	l += sprintf(l," xcalls/s:   %s |", dbl2str(s->XcallsR));
	l += sprintf(l," sysexec/s:  %s |", dbl2str(s->SysexecR));
	mvaddstr(num++,0,line);
	addstr(" scan/s:      ");
	if (s->ScanR)
		attrset(A_UNDERLINE|stress);
	addstr(dbl2str(s->ScanR));
	if (s->ScanR)
		attrset(underline);
	sprintf(line," | NfsOut/s:   %s ", dkb2str(s->NfsOR));
	addstr(line);
	if (num == Numlines) goto done;

	l = line;
	l += sprintf(l," pswitch/s:  %s |", dbl2str(s->PswitchR));
	l += sprintf(l," sysread/s:  %s |", dbl2str(s->SysreadR));
	l += sprintf(l," pgswapin/s:  %s |", dbl2str(s->PgswapinR));
	l += sprintf(l," DiskIn/s:   %s ", dkb2str(s->DiskIR));
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	l = line;
	l += sprintf(l," migrate/s:  %s |", dbl2str(s->MigrateR));
	l += sprintf(l," syswrite/s: %s |", dbl2str(s->SyswriteR));
	l += sprintf(l," pgswapout/s: %s |", dbl2str(s->PgswapoutR));
	l += sprintf(l," DiskOut/s:  %s ", dkb2str(s->DiskOR));
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	l = line;
	l += sprintf(l," invswtch/s: %s |", dbl2str(s->InvSwtchR));
	l += sprintf(l,"                   |");
	l += sprintf(l,"                    |");
	l += sprintf(l," disk errs: %5lu  ", s->HardErrors+s->SoftErrors);
	attrset(underline);
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	/*
	l = line;
	switch (s->Zpools) {
	default:
		l += sprintf(l," %d Zpools:   %s | ", s->Zpools, kb2str(s->Zsize));
		break;
	case 1:
		l += sprintf(l," %d Zpool:    %s | ", s->Zpools, kb2str(s->Zsize));
		break;
	case 0:
		l += sprintf(l," Storage:    %s | ", kb2str(s->Storage));
		break;
	}
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	attrset(underline);
	l = line;
	if (s->Zpools)
		l += sprintf(l," zfs free:   %s | ", kb2str(s->Zfree));
	else
		l += sprintf(l," FsFree:     %s | ", kb2str(s->FsFree));
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;
	*/

	ulines = Numlines - num - fslines;
	if (ulines < fslines) {
		fslines = (Numlines - num) / 2;
		ulines = Numlines - num - fslines;
		assert(ulines >= fslines);
	}
	dbug("fslines = %d\n",fslines);
	attrset(nostress);
	s->us = sort_users(s->us);
	u = s->us;
	if ((num + 1) < (Numlines - fslines)) {
		unsigned numuser = 0;
		userstat_t *us = s->us;
		while (us) {
			++numuser;
			us = us->next;
		}
		if (u->h_s) {
			if (numuser < 10)
				sprintf(line,"%d User   NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS    H+S   %%MEM  %%CPU   CPU-TIME",numuser);
			else if (numuser < 100)
				sprintf(line,"%d User  NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS    H+S   %%MEM  %%CPU   CPU-TIME",numuser);
			else 
				sprintf(line,"%d User NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS    H+S   %%MEM  %%CPU   CPU-TIME",numuser);
		} else {
			if (numuser < 10)
				sprintf(line,"%d User         NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS   %%MEM  %%CPU    CPU-TIME",numuser);
			else if (numuser < 100)
				sprintf(line,"%d User        NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS   %%MEM  %%CPU    CPU-TIME",numuser);
			else 
				sprintf(line,"%d User       NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS   %%MEM  %%CPU    CPU-TIME",numuser);
		}
		mvaddstr(num++,0,line);
	}
	while (u && (num < (Numlines-fslines))) {
		long h,m,s;
		unsigned ul = 0;
		char *l = line;
		if ((u->next == 0) || (num == (Numlines - fslines - 1))) {
			ul = A_UNDERLINE;
			attrset(nostress | A_UNDERLINE);
		}
		assert(u->time.tv_sec >= 0);
		assert(u->time.tv_nsec >= 0);
		s = u->time.tv_sec + (long) rint((double)u->time.tv_nsec/1000000000);
		m = u->time.tv_sec / 60;
		s %= 60;
		h = m / 60;
		m -= h * 60;
		assert((s >= 0) && (s < 60) && (m >= 0) && (m < 60) && (h >= 0));
		if (u->h_s) {
			sprintf(line,"%-9s %4d (%2d/%2d/",
					u->username,
					u->nproc,
					u->oncpu,
					u->sleeping);
			mvaddstr(num,0,line);
			if (u->runnable)
				attrset(stress | ul);
			sprintf(line,"%2d", u->runnable);
			addstr(line);
			attrset(nostress | ul);
			l += sprintf(l,"/%2d/%2d) %4d  %s  ",
					u->zombie,
					u->stopped,
					u->nlwp,
					kb2str(u->size));
			l += sprintf(l,"%s  ",kb2str(u->rss));
			l += sprintf(l,"%s  ",kb2str(u->h_s));
			if (u->pctmem < 1)
				l+= sprintf(l,"%.1f%%  ",u->pctmem);
			else if (u->pctmem < 10)
				l+= sprintf(l,"%3.1f%%  ",u->pctmem);
			else
				l+= sprintf(l,"%3.f%%  ",u->pctmem);
			if (u->pctcpu < 1)
				l+= sprintf(l,"%.1f%% ",u->pctcpu);
			else if (u->pctcpu < 10)
				l+= sprintf(l,"%3.1f%% ",u->pctcpu);
			else
				l+= sprintf(l,"%3.f%% ",u->pctcpu);
			l += sprintf(l,"%4d:%02d:%02d",h,m,s);
		} else {
			sprintf(line,"%-15s %4d (%2d/%2d/",
					u->username,
					u->nproc,
					u->oncpu,
					u->sleeping);
			mvaddstr(num,0,line);
			if (u->runnable)
				attrset(stress | ul);
			sprintf(line,"%2d", u->runnable);
			addstr(line);
			attrset(nostress | ul);
			l += sprintf(line,"/%2d/%2d) %4d  %s  ",
					u->zombie,
					u->stopped,
					u->nlwp,
					kb2str(u->size));
			l += sprintf(l,"%s  ",kb2str(u->rss));
			if (u->pctmem < 1)
				l+= sprintf(l,"%.1f%%  ",u->pctmem);
			else if (u->pctmem < 10)
				l+= sprintf(l,"%3.1f%%  ",u->pctmem);
			else
				l+= sprintf(l,"%3.f%%  ",u->pctmem);
			if (u->pctcpu < 1)
				l+= sprintf(l,"%.1f%%  ",u->pctcpu);
			else if (u->pctcpu < 10)
				l+= sprintf(l,"%3.1f%%  ",u->pctcpu);
			else
				l+= sprintf(l,"%3.f%%  ",u->pctcpu);
			sprintf(l,"%4d:%02d:%02d",h,m,s);
		}
		addstr(line);
		u = u->next;
		dbug("us@%d\n",num);
		++num;
	}
	fs = s->fs;
	attrset(nostress);
	while ((num < Numlines) && fs) {
		char *l = line;
		if (unshown) {
			if ((num + 1 == Numlines) || (0 == fs->next))
				attrset(underline);
		}
		if (strlen(fs->dev) > 30)
			l += sprintf(l,"%.27s...  %-4s ",fs->dev,fs->typ);
		else
			l += sprintf(l,"%-30s  %-4s ",fs->dev,fs->typ);
		if (strlen(fs->mnt) > 30)
			l += sprintf(l,"%.27s...  %s/",fs->mnt,kb2str(fs->free));
		else
			l += sprintf(l,"%-30s  %s/",fs->mnt,kb2str(fs->free));
		strcpy(l,kb2str(fs->blocks));
		mvaddstr(num,0,line);
		fs = fs->next;
		dbug("fs@%d:%s\n",num,line);
		++num;
	}
done:
	if (unshown) {
		sysstat_t *u = s->next;
		int len = 7;
		size_t maxlen = sizeof(line) - 1;
		
		attrset(nostress);
		if (Numcol < maxlen)
			maxlen = Numcol;
		++Numlines;
		mvaddstr(num, 0, "hosts: ");
		while (u != s) {
			char *d,*b;
			size_t l;

			b = strchr(u->Hostinfo,' ');
			d = strchr(u->Hostinfo,'.');
			if (b && d) {
				if (b < d)
					l = b - u->Hostinfo;
				else
					l = d - u->Hostinfo;
			} else if (b) {
				l = b - u->Hostinfo;
			} else if (d) {
				l = d - u->Hostinfo;
			} else {
				assert(0);
			}
			len += l;
			if (len >= maxlen)
				break;
			strncpy(line,u->Hostinfo,l);
			line[l] = 0;
			if (now - u->TS > 60) 
				attrset(stress);
			addstr(line);
			attrset(nostress);
			u = u->next;
			if (u != s) {
				len += 2;
				if (len >= maxlen)
					break;
				addstr(", ");
			}
		}
		if (len < Numcol) {
			memset(line,' ',Numcol - len);
			line[Numcol - len] = 0;
			addstr(line);
		}
	} else 
		--num;
	memset(line,' ',Numcol);
	line[Numcol] = 0;
	while (++num < Numlines)
		mvaddstr(num,0,line);
	move(0,0);
	refresh();
	return 1;
}

// vim:tw=0
