/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "log.h"
#include "userstat.h"

#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <procfs.h>
#include <pwd.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>


typedef struct procentr_s
{
	pid_t pid; /* six chars for pid, one for \0 */
	int psi,pst,upd;
	struct procentr_s *next, *prev;
} procentr_t;

static procentr_t *ProcEntr = 0;
static userstat_t *UserHead = 0;
size_t Nproc, Nlwp;
static int sort_nproc(userstat_t *u1, userstat_t *u2);
static int (*Crit)(userstat_t *, userstat_t *) = sort_nproc;


static userstat_t *
get_userstat(uid_t uid)
{
	userstat_t *u = UserHead;
	while ((u != 0) && (u->uid != uid))
		u = u->next;
	if (u == 0) {
		struct passwd *p;
		u = calloc(1,sizeof(userstat_t));
		assert(u);
		u->uid = uid;
		if ((p = getpwuid(uid)))
			u->username = strdup(p->pw_name);
		else
			u->username = strdup("<unknown>");
		u->next = UserHead;
		UserHead = u;
	}
	return u;
}


static void
add_procentr(procentr_t *e)
{
	procentr_t *c = ProcEntr;
	pid_t pid = e->pid;

	if (c == 0) {
		ProcEntr = e;
		e->next = 0;
		e->prev = 0;
		return;
	}
	if (c->pid < pid) {
		while (c->next && (c->next->pid < pid))
			c = c->next;
		if (c->next == 0) {
			c->next = e;
			e->prev = c;
			e->next = 0;
		} else {
			c->next->prev = e;
			e->next = c->next;
			c->next = e;
			e->prev = c;
		}
	} else {
		assert(pid != c->pid);
		while (c->prev && (c->prev->pid > pid))
			c = c->prev;
		if (c->prev == 0) {
			c->prev = e;
			e->next = c;
			e->prev = 0;
		} else {
			c->prev->next = e;
			e->prev = c->prev;
			c->prev = e;
			e->next = c;
		}
	}
	ProcEntr = e;
}


static procentr_t *
get_procentr(pid_t pid)
{
	procentr_t *c = ProcEntr;
	if (c == 0) {
		return 0;
	}
	if (c->pid == pid)
		return c;
	else if (c->pid < pid) {
		while (c && c->pid < pid)
			c = c->next;
		if (c && (c->pid == pid)) {
			ProcEntr = c;
			return c;
		}
	} else {
		while (c && c->pid > pid)
			c = c->prev;
		if (c && (c->pid == pid)) {
			ProcEntr = c;
			return c;
		}
	}
	return 0;
}


static void
rm_procentr(procentr_t *pe)
{
	if (pe->prev)
		pe->prev->next = pe->next;
	if (pe->next)
		pe->next->prev = pe->prev;
	if (pe == ProcEntr) {
		if (pe->next)
			ProcEntr = pe->next;
		else
			ProcEntr = pe->prev;
	}
	(void) close(pe->psi);
	if (pe->pst != -1)
		(void) close(pe->pst);
	free(pe);
}


static void
cleanup_proc_cache(void)
{
	procentr_t *e = ProcEntr, *p;
	assert(e);
	while (e->next)
		e = e->next;
	do {
		p = e->prev;
		if (e->upd == 0)
			rm_procentr(e);
		else
			e->upd = 0;
		e = p;
	} while (e);
}


static void 
update_procinfo(void)
{
	static DIR *procdir = 0;
	static int procfd = -1, maxfd, euid;
	struct dirent *ent;
	userstat_t *u = UserHead;

	if (0 == procdir) {
		procfd = open("/proc", O_RDONLY);
		assert(procfd != -1);
#if SOLARIS >= 9
		procdir = fdopendir(procfd);
#else
		procdir = opendir("/proc");
#endif
		assert(procdir);
		maxfd = sysconf(_SC_OPEN_MAX) - 10;
		if (euid = geteuid())
			euid = seteuid(0);
	}
	while (u) {
		u->nproc = 0;
		u->nlwp = 0;
		u->oncpu = 0;
		u->sleeping = 0;
		u->zombie = 0;
		u->runnable = 0;
		u->stopped = 0;
		u->rss = 0;
		u->size = 0;
		u->h_s = 0;
		u->pctcpu = 0;
		u->pctmem = 0;
		u->time.tv_sec = 0;
		u->time.tv_nsec = 0;
		u = u->next;
	}
	u = UserHead;
	Nproc = 0;
	Nlwp = 0;
	rewinddir(procdir);
	while (ent = readdir(procdir)) {
		int psi, pst = -1;
		psinfo_t p;
		pstatus_t s;
		userstat_t *u;
		long long ns;
		pid_t pid;
		procentr_t *pe;

		switch(ent->d_name[0]) {
		case '.':
			switch (ent->d_name[1]) {
			case '.': if (ent->d_name[2] != '\0')
					  break;
				  /*FALLTHROUGH*/
			case '\0':
				continue;
			default:;
			}
			/*FALLTHROUGH*/
		default:;
		}
		/* I think, the above is a little bit more elegant than
		 * this... ;-)
		if (0 == strcmp(".",ent->d_name))
			continue;
		if (0 == strcmp("..",ent->d_name))
			continue;
		*/
		pid = strtol(ent->d_name,0,10);
		if (pe = get_procentr(pid)) {
			psi = pe->psi;
			pst = pe->pst;
		} else {
			char psinfo[PATH_MAX];
			strcpy(psinfo,ent->d_name);
			strcat(psinfo+1,"/psinfo");
#if SOLARIS >= 9
			psi = openat(procfd,psinfo,O_RDONLY);
			if (0 == euid) {
				strcpy(psinfo,ent->d_name);
				strcat(psinfo+1,"/status");
				pst = openat(procfd,psinfo,O_RDONLY);
			}
#else
			psi = open(psinfo,O_RDONLY);
			if (0 == euid) {
				strcpy(psinfo,ent->d_name);
				strcat(psinfo+1,"/status");
				pst = open(psinfo,O_RDONLY);
			}
#endif
			if ((-1 == psi) || ((0 == euid) && (-1 == pst))) {
				/* process died after reading directory entry */
				if (psi != -1)
					(void) close(psi);
				if (errno == EMFILE)
					error("ran out of file descriptors\n");
				continue;
			}
			pe = malloc(sizeof(procentr_t));
			pe->psi = psi;
			pe->pst = pst;
			pe->pid = pid;
			add_procentr(pe);
		}
		if ((pread(psi,&p,sizeof(p),0) != sizeof(p)) || ((euid == 0) && (pread(pst,&s,sizeof(s),0) != sizeof(s)))) {
			/* process died after opening the file */
			rm_procentr(pe);
			continue;
		}
		if (psi >= maxfd)
			rm_procentr(pe);
		pe->upd = 1;
		++Nproc;
		Nlwp += p.pr_nlwp;
		u = get_userstat(p.pr_uid);
		++u->nproc;
		u->nlwp += p.pr_nlwp;
		u->size += p.pr_size;
		u->rss += p.pr_rssize;
		if (euid == 0)
			u->h_s += (s.pr_stksize + s.pr_brksize)/1024;
		u->pctcpu += (double)p.pr_pctcpu/0x8000*100;
		u->pctmem += (double)p.pr_pctmem/0x8000*100;
		u->time.tv_sec += p.pr_time.tv_sec;
		u->time.tv_sec += p.pr_ctime.tv_sec;
		ns = u->time.tv_nsec + p.pr_time.tv_nsec;
		u->time.tv_sec += ns / 1000000000;
		u->time.tv_nsec = ns % 1000000000;
		ns += p.pr_ctime.tv_nsec;
		u->time.tv_sec += ns / 1000000000;
		u->time.tv_nsec = ns % 1000000000;
		assert((u->time.tv_sec >= 0) && (u->time.tv_nsec >= 0));
		switch(p.pr_lwp.pr_sname) {
		default:
			abort();
			break;
		case 'O':
			++u->oncpu;
			break;
		case 'S':
			++u->sleeping;
			break;
		case 'R':
			++u->runnable;
			break;
		case 'Z':
			++u->zombie;
			break;
		case 'T':
			++u->stopped;
			break;
		}
	}
	cleanup_proc_cache();
}


static int 
sort_pcpu(userstat_t *u1, userstat_t *u2)
{
	return u1->pctcpu < u2->pctcpu;
}


static int 
sort_nproc(userstat_t *u1, userstat_t *u2)
{
	return u1->nproc < u2->nproc;
}


static int 
sort_rss(userstat_t *u1, userstat_t *u2)
{
	return u1->rss < u2->rss;
}


static int 
sort_size(userstat_t *u1, userstat_t *u2)
{
	return u1->size < u2->size;
}


static int 
sort_hs(userstat_t *u1, userstat_t *u2)
{
	if (u1->h_s)
		return u1->h_s < u2->h_s;
	return u1->rss < u2->rss;
}


static int 
sort_time(userstat_t *u1, userstat_t *u2)
{
	return u1->time.tv_sec < u2->time.tv_sec;
}


static int 
sort_user(userstat_t *u1, userstat_t *u2)
{
	return strcmp(u1->username,u2->username) > 0;
}


void 
set_sort_by_pcpu(void)
{
	Crit = sort_pcpu;
}


void 
set_sort_by_nproc(void)
{
	Crit = sort_nproc;
}


void 
set_sort_by_rss(void)
{
	Crit = sort_rss;
}


void
set_sort_by_size(void)
{
	Crit = sort_size;
}


void
set_sort_by_time(void)
{
	Crit = sort_time;
}


void
set_sort_by_hs(void)
{
	Crit = sort_hs;
}


void
set_sort_by_user(void)
{
	Crit = sort_user;
}


userstat_t *
sort_users(userstat_t *u)
{
	userstat_t *h = 0, *t = 0;
	while (u) {
		userstat_t *n = u->next, *i = h;
		if (u->nproc == 0) {
			free(u->username);
			free(u);
			u = n;
			continue;
		}
		while (i && Crit(u,i)
			&& i->next && Crit(u,i->next)) {
			assert(i != i->next);
			i = i->next;
		}
		if (i == 0) {
			if (t) {
				assert(h && !Crit(t, i));
				t->next = u;
				t = u;
			} else {
				assert(h == 0);
				h = t = u;
			}
			u->next = 0;
		} else {
			if (Crit(u,i)) {
				u->next = i->next;
				i->next = u;
			} else {
				u->next = i;
				h = u;
			}
			assert(i != i->next);
		}
		u = n;
		assert(u == 0 || u != u->next);
	}
	return h;
}


userstat_t *
get_procinfo(void)
{
	update_procinfo();
	UserHead = sort_users(UserHead);
	return UserHead;
}

// vim:tw=0
